"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, List
from kudubot.users.Contact import Contact
from kudubot.entities.Message import Message
from kudubot.services.BaseService import BaseService


# noinspection PyAbstractClass
class MultiLanguageService(BaseService):
    """
    Service Extension class that enables support for multiple languages
    """

    def determine_language(self, message: Message) -> str:
        """
        Checks a message object for any language indicators to determine
        in which language to reply with

        :param message: The message to analyze
        :return: The language key
        """
        raise NotImplementedError()

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        Defines the dictionary with which the text is translated
        in the translate() method.
        This should be in the form of a dictionary like this:

        { key: {"language": "text_in_language", ...}, ... }

        Keep in mind that every instance of the 'key'
        value is replaced while translating

        :return: The dictionary to create translations with
        """
        raise NotImplementedError

    # noinspection PyMethodMayBeStatic
    def define_fallback_language(self) -> str:
        """
        Defines a fallback language in case a language
        is not implemented for a key.

        :return: By default, the language "en" is returned
        """
        return "en"

    # ----------------------------------------------------------------------- #
    #                  _,-'/-'/                              Here be dragons! #
    #  .      __,-; ,'( '/                                                    #
    #   \.    `-.__`-._`:_,-._       _ , . ``                                 #
    #    `:-._,------' ` _,`--` -: `_ , ` ,' :                                #
    #       `---..__,,--'            ` -'. -'                                 #
    # Everything below this should not be overridden by subclasses            #
    # -------------------------------------------------------------------------

    lang_switch_command_keywords = ["/language", "/sprache"]
    """
    List of command triggers that can be used to list or change the language
    """

    lang_switch_aliases = {
        "en": ["en", "english", "englisch"],
        "de": ["de", "deutsch", "german"]
    }
    """
    Aliases for language names in other languages
    """

    def supported_languages(self) -> List[str]:
        """
        :return: A list of languages supported by the Service
        """
        supported = []

        dictionary = self.define_language_text()
        for token in dictionary:
            for language in dictionary[token]:
                if language not in supported:
                    supported.append(language)
        return supported

    def translate(self, text: str, language: str,
                  translation_dict: Dict[str, Dict[str, str]] = None) -> str:
        """
        Translates text using the service's dictionary
        in the specified language

        :param text: The text to translate
        :param language: The language to translate into
        :param translation_dict: Can be specified to
                                 determine a custom dictionary
        :return: The translated text
        """

        translated = text
        if translation_dict is None:
            language_text = self.define_language_text()
        else:
            language_text = dict(translation_dict)

        for key in language_text:
            try:
                translated = translated.replace(
                    key, language_text[key][language]
                )
            except KeyError:
                try:
                    translated = translated.replace(
                        key,
                        language_text[key][self.define_fallback_language()]
                    )
                except KeyError:  # No English language defined
                    return text

        return translated

    def reply_translated(self, title: str, body: str, message: Message):
        """
        Provides a helper method that streamlines the process
        of replying to a message.
        Very useful for Services that send a reply immediately
        to cut down on clutter in the code

        In addition to the standard reply method,
        this method translates the text prior to sending it

        :param title: The title of the message to send
        :param body: The body of the message to send
        :param message: The message to reply to
        :return: None
        """
        language = self.determine_language(message)
        self.reply(
            self.translate(title, language),
            self.translate(body, language),
            message
        )

    # noinspection PyMethodMayBeStatic
    def is_applicable_to_multi_language(self, message: Message):
        """
        Checks if a message is applicable for a language list or change
        :param message:
        :return:
        """

        body = message.message_body.lower().strip()
        for keyword in self.lang_switch_command_keywords:
            if body.startswith(keyword + " ") or body == keyword:
                return True
        return False

    def handle_message_multi_language(self, message: Message):
        """
        Analyzes a message for the language used and stores that language
        value in the database as a preference of the user.
        Also implements the /language command which allows a user
        to view and change the current language

        :param message: The message to analyze
        :return: None
        """

        lang_switch_dict = {
            "@{title}": {
                "en": "Language Change",
                "de": "Sprachwechsel"
            },
            "@{success_message}": {
                "en": "Successfully changed language to",
                "de": "Sprache erfolgreich geändert zu"
            },
            "@{fail_message}": {
                "en": "Failed to switch to language",
                "de": "Konnte nicht Sprache wechseln zu"
            }
        }

        params = message.message_body.lower().split(" ")
        contact = message.get_direct_response_contact()

        # Example Message: '/language'
        if len(params) == 1 and params[0] in self.lang_switch_command_keywords:
            language = self.get_stored_language_preference(contact)

            self.reply(self.translate("@{title}", language, lang_switch_dict),
                       language, message)

        # Example Message: '/language english'
        elif len(params) == 2 \
                and params[0] in self.lang_switch_command_keywords:

            found_language = False
            for key in self.lang_switch_aliases:

                for alias in self.lang_switch_aliases[key]:

                    if alias == params[1]:
                        found_language = True
                        self.connection.language_selector.\
                            store_language_preference(contact, key, True)
                        break

            language = self.get_stored_language_preference(contact)
            # the new language will already be stored
            title = self.translate("@{title}", language, lang_switch_dict)

            if found_language:
                self.reply(title, self.translate("@{success_message}: " +
                                                 language, language,
                                                 lang_switch_dict), message)
            else:
                self.reply(title, self.translate("@{fail_message}: " +
                                                 params[1], language,
                                                 lang_switch_dict), message)

        else:
            language = self.determine_language(message)
            self.connection.language_selector.store_language_preference(
                message.get_direct_response_contact(), language
            )

    def get_stored_language_preference(self, contact: Contact) -> str:
        """
        Gets the stored language preference of a contact
        :param contact: The contact to get the stored language preference for
        :return: The language
        """
        return self.connection.language_selector.get_language_preference(
            contact, "en"
        )



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import datetime
import re
import time
from typing import Dict

from kudubot.entities.Message import Message
from kudubot.services.HelperService import HelperService
from kudubot.services.internal.reminder.database import initialize_database, \
    store_reminder, get_unsent_reminders, \
    mark_reminder_sent, convert_datetime_to_string


class ReminderService(HelperService):
    """
    Class that implements a Service for the Kudubot framework that allows
    users to store reminder message that are then sent at a later time
    """

    def init(self):
        """
        Initializes the database table and starts a background thread that
        perpetually searches for expired reminders
        """
        self.initialize_database_table(initializer=initialize_database)
        self.start_daemon_thread(self.background_loop)

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the identifier for this service

        :return: The Service's identifier
        """
        return "reminder"

    def define_command_name(self, language: str):
        """
        Defines the command name for this service
        :param language: the language in which to get the command name
        :return: The command name in the specified language
        """
        return {"en": "/remind", "de": "/erinner"}[language]

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        :return: A dictionary used to translate any user-facing messages
        """
        return {
            "@remind_command": {"en": "/remind", "de": "/erinner"},
            "@list_argument": {"en": "list", "de": "auflisten"},
            "@second_singular": {"en": "second", "de": "sekunde"},
            "@second_plural": {"en": "seconds", "de": "sekunden"},
            "@minute_singular": {"en": "minute", "de": "minute"},
            "@minute_plural": {"en": "minutes", "de": "minuten"},
            "@hour_singular": {"en": "hour", "de": "stunde"},
            "@hour_plural": {"en": "hours", "de": "stunden"},
            "@day_singular": {"en": "day", "de": "tag"},
            "@day_plural": {"en": "days", "de": "tage"},
            "@week_singular": {"en": "week", "de": "woche"},
            "@week_plural": {"en": "weeks", "de": "wochen"},
            "@year_singular": {"en": "year", "de": "jahr"},
            "@year_plural": {"en": "years", "de": "jahre"},
            "@stored_reply_title": {
                "en": "Message Stored", "de": "Nachricht gespeichert"
            },
            "@stored_confirmation_message": {
                "en": "The reminder message has successfully been stored.",
                "de": "Die Erinnerungsnachricht wurde erfolgreich gespeichert"
            },
            "@list_stored_message_start": {
                "en": "The following reminders are pending",
                "de": "Die folgenden Erinnerungen stehen noch aus"
            },
            "@list_response_title": {
                "en": "List of Reminders", "de": "Liste der Erinnerungen"
            },
            "@invalid_title": {
                "en": "Invalid Reminder", "de": "Ungültige Erinnerung"
            },
            "@invalid_message": {
                "en": "The reminder is invalid. "
                      "Please consult /remind syntax for more information.",
                "de": "Die Erinnerung ist fehlerhaft."
                      "Sehe dir die /erinner syntax an, "
                      "und versuch es dann noch einmal."
            }
        }

    def handle_message(self, message: Message):
        """
        Handles a message received by the Service

        :param message: The message to handle
        :return: None
        """
        target = message.get_direct_response_contact()
        command = self.parse_message(message.message_body.strip(),
                                     self.determine_language(message))

        if command["mode"] == "store":
            store_reminder(self.connection.db,
                           command["data"]["message"],
                           command["data"]["due_time"],
                           target.database_id)
            self.reply_translated("@stored_reply_title",
                                  "@stored_confirmation_message", message)

        elif command["mode"] == "list":
            reminders = list(filter(
                lambda x: x["receiver"].database_id == target.database_id,
                get_unsent_reminders(self.connection.db)
            ))
            text = "@list_stored_message_start:\n\n"
            for reminder in reminders:
                text += convert_datetime_to_string(reminder["due_time"])
                text += ":" + reminder["message"] + "\n"
            self.reply_translated("@list_response_title", text, message)

        elif command["mode"] == "invalid":
            self.reply_translated(
                "@invalid_title", "@invalid_message", message)

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if the Service is applicable to a message

        :param message: The message to check
        :return: True if the Service is applicable, otherwise False
        """
        language = self.determine_language(message)
        body = message.message_body.lower().strip()

        regex = "^@remind_command ([0-9]+ (" \
                "@second_singular|@second_plural|@minute_singular|" \
                "@minute_plural|@hour_singular|@hour_plural|" \
                "@day_singular|@day_plural|@week_singular|@week_plural" \
                "|@year_singular|@year_plural) )+" \
                "\"[^\"]+\"$"
        lang_regex = re.compile(self.translate(regex, language))

        return re.match(lang_regex, body) \
            or body == self.translate(
                "@remind_command @list_argument", language)

    def background_loop(self):
        """
        Perpetually checks for expiring reminders.

        :return: None
        """
        db = self.connection.get_database_connection_copy()
        while True:
            unsent = get_unsent_reminders(db)

            for reminder in unsent:

                due_time = reminder["due_time"].timestamp()
                now = datetime.datetime.utcnow().timestamp()
                if due_time < now:

                    self.connection.send_message(Message(
                        "Reminder",
                        reminder["message"],
                        reminder["receiver"],
                        self.connection.user_contact)
                    )
                    mark_reminder_sent(db, reminder["id"])

            time.sleep(1)

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the syntax with which the user can interact with this service

        :param language: The language to use
        :return: The syntax description in the specified language
        """
        return {
            "en": "Set a reminder:\n"
                  "/remind X second(s) \"Message\"\n"
                  "/remind X minute(s) \"Message\"\n"
                  "/remind X hour(s) \"Message\"\n"
                  "/remind X day(s) \"Message\"\n"
                  "/remind X week(s) \"Message\"\n"
                  "/remind X year(s) \"Message\"\n"
                  "/remind tomorrow \"Message\"\n"
                  "/remind next week \"Message\"\n"
                  "/remind next year \"Message\"\n\n"
                  "Combinations:\n"
                  "/remind X hours Y minutes Z seconds \"Message\"\n\n"
                  "List pending reminders:\n"
                  "/remind list",
            "de": "Erinnerung setzen:\n"
                  "/erinner X sekunde(n) \"Message\"\n"
                  "/erinner X minute(n) \"Message\"\n"
                  "/erinner X stunde(n) \"Message\"\n"
                  "/erinner X tag(e) \"Message\"\n"
                  "/erinner X woche(n) \"Message\"\n"
                  "/erinner X jahr(e) \"Message\"\n"
                  "/erinner morgen \"Message\"\n"
                  "/erinner nächste woche \"Message\"\n"
                  "/erinner nächstes jahr \"Message\"\n\n"
                  "Kombinationen:\n"
                  "/erinner X stunden Y minuten Z sekunden \"Message\"\n\n"
                  "Gespeicherte Erinnerungen auflisten:\n"
                  "/erinner auflisten"
        }[language]

    def determine_language(self, message: Message) -> str:
        """
        Determines the language used in a message

        :param message: The message to analyse
        :return: The language that was found
        """

        if message.message_body.startswith("/erinner"):
            return "de"
        else:
            return "en"

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message for this service in various languages

        :param language: The language to be used
        :return: The help description in the specified language
        """
        return {
            "en": "The reminder service allows you to store a reminder "
                  "on the server to be sent back to you at a specified time. "
                  "See /remind syntax for possible ways "
                  "to use the reminder service.",
            "de": "Der Erinnerungsdienst erlaubt es einem Nutzer eine "
                  "Erinnerung auf dem Server zu speichern, welcher dann zu"
                  "einem spezifizierten Zeitpunkt zurückgesendet wird. "
                  "Sehe dir /erinner syntax an, um dir ein Bild davon zu "
                  "machen wie man den Erinnerungsdienst verwendet."
        }[language]

    def parse_message(self, text: str, language: str) \
            -> Dict[str, str or Dict[str, str or datetime]]:
        """
        Parses the message and determines the mode of operation

        :param text: The text to parse
        :param language: The language in which to parse the text
        :return: A dictionary with at least the key 'status' with three
                 different possible states:
                 - no-match: The message does not match the command syntax
                 - help:     A query for the help message.
                             Will result in the help message being
                             sent to the sender
                 - store:    Command to store a new reminder
        """
        self.logger.debug("Parsing message")

        if self.translate("@remind_command @list_argument", language) \
                == text.lower():
            return {"mode": "list"}
        else:

            time_string = text.lower().split(" \"")[0].split(
                self.translate("@remind_command ", language))[1]
            reminder_message = text.split("\"")[1]

            usertime = self.parse_time_string(time_string.strip(), language)

            if usertime is None:
                self.logger.debug("Invalid reminder message")
                return {"mode": "invalid"}
            else:
                self.logger.debug("Will store reminder")
                return {
                    "mode": "store",
                    "data": {
                        "message": reminder_message,
                        "due_time": usertime
                    }
                }

    def parse_time_string(self, time_string: str, language: str) \
            -> datetime or None:
        """
        Parses a time string like '1 week' or '2 weeks 1 day' etc. and returns
        a datetime object with the specified time difference to the current
        time.

        :param time_string: The time string to parse
        :param language: In which language the string should be parsed
        :return: The parsed datetime object or None in case the parsing failed
        """

        # turn all units into english singular units
        for unit in ["second", "minute", "hour", "day", "week", "year"]:

            # The order is very important here!
            for mode in ["plural", "singular"]:

                key = "@" + unit + "_" + mode
                text = self.translate(key, language)
                time_string = time_string.replace(text, unit)

        now = datetime.datetime.utcnow()
        parsed = time_string.split(" ")

        try:
            for i in range(0, len(parsed), 2):

                value = int(parsed[i])
                key = parsed[i + 1]

                if key == "year":
                    now += datetime.timedelta(days=value * 365)
                elif key == "week":
                    now += datetime.timedelta(weeks=value)
                elif key == "day":
                    now += datetime.timedelta(days=value)
                elif key == "hour":
                    now += datetime.timedelta(hours=value)
                elif key == "minute":
                    now += datetime.timedelta(minutes=value)
                elif key == "second":
                    now += datetime.timedelta(seconds=value)
                else:
                    self.logger.debug("Invalid time keyword " + key + " used")
                    return None

            return now

        # Datetime exception, too high date values
        except (ValueError, OverflowError):
            return None



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from kudubot.entities.Message import Message
from kudubot.services.MultiLanguageService import MultiLanguageService


# noinspection PyAbstractClass
class HelperService(MultiLanguageService):
    """
    Service extension that allows for the automatic sending of
    help and syntax messages. Provides support for multiple languages
    """

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message for the Service

        :param language: The language in which to get the help message in
        :return: The help message in the specified language
        """
        raise NotImplementedError()

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the syntax description for this service.

        :param language: The language in which to get the syntax description in
        :return: The syntax description in the specified language
        """
        raise NotImplementedError()

    # noinspection PyUnusedLocal
    def define_command_name(self, language: str) -> str:
        """
        Defines the command name used to call this Service

        :param language: The language for the command name,
                         for supporting different command names for
                         different languages
        :return: The command name for this service.
                 Defaults to a forward slash and the Service's identifier
        """
        return "/" + self.define_identifier()

    # -------------------------------------------------------------------------
    #                  _,-'/-'/                                Here be dragons!
    #  .      __,-; ,'( '/
    #   \.    `-.__`-._`:_,-._       _ , . ``
    #    `:-._,------' ` _,`--` -: `_ , ` ,' :
    #       `---..__,,--'            ` -'. -'
    # Everything below this should not be overridden by subclasses
    # -------------------------------------------------------------------------

    def handle_message_helper(self, message: Message):
        """
        Handles the help message sending.
        Checks if a message qualifies for a help message and then sends
        messages accordingly.

        :param message: The message to handle
        :return: None
        """
        body = message.message_body

        dictionary = {
            "@help_message_title":
                {"en": "Help Message for " + self.identifier,
                 "de": "Hilfnachricht für " + self.identifier},
            "@syntax_message_title":
                {"en": "Syntax Message for " + self.identifier,
                 "de": "Syntaxnachricht für " + self.identifier}
        }
        help_keywords = ["help", "hilfe"]
        syntax_keywords = ["syntax"]

        try:
            language = self.determine_language(message)
        except NotImplementedError:
            language = self.define_fallback_language()
        command_keyword = self.define_command_name(language)

        # Checks if the /help or /syntax commands were used
        # If that's the case, adjust the message text.
        for keyword in help_keywords + syntax_keywords:
            if body.lower().split(" ")[0] == "/" + keyword:
                for service in self.connection.services:
                    identifier = body.lower().split(" ")[1].strip()
                    if service.identifier == identifier:
                        if service.identifier != self.identifier:
                            return
                        else:
                            body = command_keyword + " " + keyword

        if body.startswith(self.define_command_name(language)):
            body = body.split(self.define_command_name(language), 1)[1].strip()

            if body in help_keywords:

                title = self.translate("@help_message_title",
                                       language, dictionary)
                try:
                    body = self.define_help_message(language)
                except KeyError:
                    body = "NOT_DEFINED_FOR_LANG"
                    self.logger.error("Missing language definition: " +
                                      language)

                self.reply(title, body, message)

            elif body in syntax_keywords:

                title = self.translate("@syntax_message_title",
                                       language, dictionary)
                try:
                    body = self.define_syntax_description(language)
                except KeyError:
                    body = "NOT_DEFINED_FOR_LANG"
                    self.logger.error("Missing language definition: " +
                                      language)

                self.reply(title, body, message)

    def is_applicable_to_helper(self, message: Message) -> bool:
        """
        Checks if the message is applicable to the service by checking
        if the command name is followed by the terms 'help' or 'syntax'.

        :param message: The message to analyze
        :return: True if the message is applicable, False otherwise
        """
        language = self.determine_language(message)
        command = self.define_command_name(language).lower()
        params = message.message_body.lower().split(" ")
        identifier = self.identifier

        # Needs at least 2 parameters. Command or identifier + help
        if len(params) < 2:
            return False

        dictionary = {
            "@help_command": {"en": "help",
                              "de": "hilfe"},
            "@syntax_command": {"en": "syntax",
                                "de": "syntax"}
        }

        # Check if help or syntax keyword in message text.
        # Change language accordingly
        for entry, values in dictionary.items():
            for lang, keyword in values.items():
                if params[0] == "/" + keyword:
                    language = lang
                elif params[1] == keyword:
                    language = lang

        help_command = self.translate("@help_command", language, dictionary)
        syntax_command = self.translate("@syntax_command", language,
                                        dictionary)

        if len(params) == 2 and "/" + help_command == params[0] \
                and params[1] == identifier:
            # /help identifier
            return True

        elif len(params) == 2 and "/" + syntax_command == params[0] \
                and params[1] == identifier:
            # /syntax identifier
            return True

        else:
            # /command help
            return message.message_body.lower() in [
                command + " " + help_command,
                command + " " + syntax_command
            ]

    def starts_with_command_keyword(self, message: Message, language: str,
                                    case_sensitive: bool = False) -> bool:
        """
        Checks if a message text starts with the command keyword
        defined by define_command_name()

        :param message: The message to check
        :param language: The language for which to check
        :param case_sensitive: Can be set to True to do a case-sensitive check
        :return: True if the message starts with the command name, else False
        """
        if case_sensitive:
            return message.message_body.\
                startswith(self.define_command_name(language))

        else:
            return message.message_body.lower(). \
                startswith(self.define_command_name(language).lower())



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import unittest
from kudubot.users.Contact import Contact
from kudubot.tests.helpers.DummyConnection import DummyConnection
from kudubot.tests.helpers.test_config import generate_test_environment,\
    clean_up_test_environment


# noinspection SqlNoDataSourceInspection
class UnitTests(unittest.TestCase):
    """
    Class that tests the AddressBook class
    """

    def setUp(self):
        """
        Changes the class variables for testing and
        creates the testing directory

        :return: None
        """
        self.config_handler = generate_test_environment()
        self.connection = DummyConnection([], self.config_handler)
        self.connection.db.execute("DELETE FROM address_book")
        self.connection.db.commit()

    def tearDown(self):
        """
        Restores the class variables and deletes the testing directory

        :return: None
        """
        clean_up_test_environment()

    def test_invalid_contact_fetches(self):
        """
        Tests if the contact fetch methods return a None object
        if they fail to find a result

        :return: None
        """
        self.assertEqual(
            None,
            self.connection.address_book.get_contact_for_address("No_Address")
        )
        self.assertEqual(
            None, self.connection.address_book.get_contact_for_id(100)
        )

    def test_contact_operations(self):
        """
        Tests various contact operations in the addressbook.
        Uses subtest methods to make the tests a bit more readable

        :return: None
        """
        self.subtest_updating_contact()
        contact = self.connection.address_book.add_or_update_contact(
            Contact(-1, "ABC", "DEF")
        )
        self.assertEqual(contact.display_name, "ABC")
        self.assertEqual(contact.database_id, 1)
        self.assertEqual(contact.address, "DEF")
        self.subtest_fetching_contacts()

    def subtest_fetching_contacts(self):
        """
        Tests fetching the contact information from the database
        To be used at the end of the test_contact_operations method

        :return: None
        """
        address_contact = \
            self.connection.address_book.get_contact_for_address("DEF")
        id_contact = self.connection.address_book.get_contact_for_id(1)
        self.assertEqual(id_contact.display_name, address_contact.display_name)
        self.assertEqual(id_contact.database_id, address_contact.database_id)
        self.assertEqual(id_contact.address, address_contact.address)

    def subtest_updating_contact(self):
        """
        Tests Updating a contact

        :return: None
        """

        contact = Contact(-1, "A", "B")
        self.subtest_adding_contact_to_addressbook(contact)

        new_contact = Contact(1, "ABC", "DEF")
        self.connection.address_book.add_or_update_contact(new_contact)

        self.assertEqual(
            self.connection.address_book.get_contact_for_address("B"),
            None
        )
        new_inserted = \
            self.connection.address_book.get_contact_for_address("DEF")

        self.assertEqual(new_inserted.display_name, "ABC")
        self.assertEqual(new_inserted.database_id, contact.database_id)
        self.assertEqual(new_inserted.address, "DEF")

    def subtest_adding_contact_to_addressbook(self, contact: Contact):
        """
        Tests adding a contact to the addressbook

        :param contact: The contact to add
        :return: None
        """
        self.assertEqual(-1, contact.database_id)

        result = self.connection.address_book.add_or_update_contact(contact)

        self.assertEqual(1, result.database_id)

        self.assertEqual(contact.display_name, result.display_name)
        self.assertEqual(contact.address, result.address)

        inserted = self.connection.address_book.get_contact_for_address("B")
        self.assertEqual(inserted.display_name, result.display_name)
        self.assertEqual(inserted.database_id, result.database_id)
        self.assertEqual(inserted.address, result.address)



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from kudubot.entities.Message import Message
from kudubot.services.BaseService import BaseService


# noinspection PyAbstractClass
class AuthenticatedService(BaseService):
    """
    Service that can only be used by admin users
    """

    # -------------------------------------------------------------------------
    #                  _,-'/-'/                                Here be dragons!
    #  .      __,-; ,'( '/
    #   \.    `-.__`-._`:_,-._       _ , . ``
    #    `:-._,------' ` _,`--` -: `_ , ` ,' :
    #       `---..__,,--'            ` -'. -'
    # Everything below this should not be overridden by subclasses
    # -------------------------------------------------------------------------

    def is_applicable_to_authenticated(self, message: Message) -> bool:
        """
        Checks if the sender of the message has administrative privileges
        :param message: The message to check
        :return: True if the sender is an admin, False otherwise
        """
        is_admin = self.connection.authenticator.is_admin(message.sender)
        if not is_admin:
            self.logger.info("Authorizaton denied for " +
                             message.sender.address)
        return is_admin



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
import unittest
from kudubot.connections.Connection import Connection
from kudubot.entities.Message import Message
from kudubot.tests.helpers.DummyConnection import DummyConnection
from kudubot.tests.helpers.DummyService import DummyService
from kudubot.tests.helpers.test_config import generate_test_environment, \
    clean_up_test_environment
from kudubot.users.AddressBook import Contact


class UnitTests(unittest.TestCase):
    """
    Class that tests the Connection class
    """

    def setUp(self):
        """
        :return: None
        """
        self.config_handler = generate_test_environment()
        self.connection = DummyConnection([], self.config_handler)

    def tearDown(self):
        """
        :return: None
        """
        clean_up_test_environment()

    def test_abstract_methods(self):
        """
        Tests if the methods of the connection class are abstract
        :return: None
        """
        for method in [(Connection.define_user_contact, 0),
                       (Connection.define_user_contact, 0),
                       (Connection.listen, 0),
                       (Connection.send_message, 1),
                       (Connection.send_image_message, 3),
                       (Connection.send_audio_message, 3),
                       (Connection.send_video_message, 3),
                       (Connection.generate_configuration, 0),
                       (Connection.load_config, 0),
                       (Connection.define_identifier, -1)]:
            try:

                # The self.connection arguments emulate the self parameter
                if method[1] == -1:
                    method[0]()
                if method[1] == 0:
                    method[0](self.connection)
                elif method[1] == 1:
                    method[0](self.connection, None)
                elif method[1] == 2:
                    method[0](self.connection, None, None)
                elif method[1] == 3:
                    method[0](self.connection, None, None, None)
                self.fail()
            except NotImplementedError:
                pass

    def test_daemon_thread_start(self):
        """
        Tests if the daemon thread is started correctly
        :return: None
        """
        class ThreadLessDummyConnection(DummyConnection):
            def listen(self):
                time.sleep(1)

        t = ThreadLessDummyConnection(
            [], self.config_handler).listen_in_separate_thread()

        self.assertTrue(t.is_alive())
        while t.is_alive():
            pass
        self.assertFalse(t.is_alive())

    # noinspection PyMethodMayBeStatic
    def test_processing_message(self):
        """
        Tests if the connection correctly processes a message
        using the services
        :return: None
        """

        class AlwaysApplicableDummyService(DummyService):
            def is_applicable_to(self, message: Message):
                return True

        user = Contact(1, "1", "1")

        connection = DummyConnection([DummyService], self.config_handler)
        connection.apply_services(Message("Test", "Body", user, user))

        connection = DummyConnection([AlwaysApplicableDummyService],
                                     self.config_handler)
        connection.apply_services(Message("Test", "Body", user, user), True)
        connection.apply_services(Message("Test", "Body", user, user), False)

        connection.db.close()



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import sqlite3
import logging
from kudubot.users.Contact import Contact


# noinspection SqlDialectInspection,SqlNoDataSourceInspection,SqlResolve
class Authenticator(object):
    """
    Class that handles the authentication of users
    Offers the identification of users as either admin or blacklisted
    """

    logger = logging.getLogger(__name__)
    """
    The Logger for this class
    """

    def __init__(self, db: sqlite3.Connection):
        self.db = db
        self.db.execute(
            "CREATE TABLE IF NOT EXISTS authenticator ("
            "    user_id INTEGER CONSTRAINT constraint_name PRIMARY KEY,"
            "    is_admin INTEGER NOT NULL,"
            "    is_blacklisted INTEGER NOT NULL"
            ")"
        )
        self.db.commit()
        self.logger.info("Authenticator initialized")

    def make_admin(self, contact: Contact or int,
                   database_override: sqlite3.Connection = None):
        """
        Grants admin privileges to a contact
        :param contact: The contact to grant admin privileges to
        :param database_override: Provides a different database connection.
                                  Required for use in other threads.
        """
        user_id = contact if isinstance(contact, int) else contact.database_id
        db = self.db if database_override is None else database_override
        db.execute(
            "INSERT OR REPLACE INTO authenticator "
            "(user_id, is_admin, is_blacklisted) VALUES (?, ?, ?)",
            (user_id, True, False)
        )

    def blacklist(self, contact: Contact or int,
                  database_override: sqlite3.Connection = None):
        """
        Blacklists a contact
        :param contact: The contact to blacklist
        :param database_override: Provides a different database connection.
                                  Required for use in other threads.
        """
        user_id = contact if isinstance(contact, int) else contact.database_id
        db = self.db if database_override is None else database_override
        db.execute(
            "INSERT OR REPLACE INTO authenticator "
            "(user_id, is_admin, is_blacklisted) VALUES (?, ?, ?)",
            (user_id, False, True)
        )

    def is_admin(self, contact: Contact or int,
                 database_override: sqlite3.Connection = None) -> bool:
        """
        Checks if a contact has admin privileges
        :param contact: The contact to check
        :param database_override: Provides a different database connection.
                                  Required for use in other threads.
        :return: True if the contact has admin privileges else False
        """
        if contact is None:
            return False

        user_id = contact if isinstance(contact, int) else contact.database_id
        db = self.db if database_override is None else database_override
        result = db.execute(
            "SELECT * FROM authenticator WHERE user_id=? AND is_admin=1",
            (user_id,)).fetchall()
        return len(result) == 1

    def is_blacklisted(self, contact: Contact or int,
                       database_override: sqlite3.Connection = None) -> bool:
        """
        Checks if a contact is blacklisted
        :param contact: The contact to check
        :param database_override: Provides a different database connection.
                                  Required for use in other threads.
        :return: True if the contact is blacklisted else False
        """
        if contact is None:
            return False

        user_id = contact if isinstance(contact, int) else contact.database_id
        db = self.db if database_override is None else database_override
        result = db.execute(
            "SELECT * FROM authenticator WHERE user_id=? AND is_blacklisted=1",
            (user_id,)).fetchall()
        return len(result) == 1



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import Dict
from kudubot.users.Contact import Contact
from kudubot.users.Contact import from_dict as contact_from_dict


class Message(object):
    """
    Class that models a message entity.
    """

    def __init__(self, message_title: str, message_body: str,
                 receiver: Contact, sender: Contact,
                 sender_group: Contact = None, timestamp: float = -1.0):
        """
        Initializes the Message object

        Since Messages may also come from chat groups,
        an optional group argument exists.
        A group is another contact, just like any other contact.
        Group messages must define a contact for the group as well as for the
        individual sender of the message.

        A Message object whose group is None is a private message.

        :param message_title: The title of the message. Not always applicable,
                              depending on the connection
        :param message_body: The body of the message
        :param receiver: The recipient of the message,
                         which is a Contact object
        :param sender: The sender of the message, which is a Contact object
        :param sender_group: Optionally the group Contact object if this is a
                             message originating from a group
        :param timestamp: The timestamp of the message.
                          If it is not specified,
                          the current UNIX timestamp is used
        """
        self.message_title = message_title
        self.message_body = message_body
        self.receiver = receiver
        self.sender = sender
        self.sender_group = sender_group
        self.timestamp = timestamp if timestamp != -1.0 else time.time()

    def get_direct_response_contact(self) -> Contact:
        """
        :return: The contact to which one can immediately respond to.
                 I.e. if the message came from a group,
                 this method will return the group contact,
                 private chats however will return the individual sender
        """
        return self.sender if self.sender_group is None else self.sender_group

    def to_dict(self) -> Dict[str, str or float or Dict[str, str or int]]:
        """
        :return: The message data as a dictionary which can be used
                 to store a message as a JSON file
        """

        sender_group = \
            None if self.sender_group is None else self.sender_group.to_dict()
        return {
            "message_title": self.message_title,
            "message_body": self.message_body,
            "sender": self.sender.to_dict(),
            "sender_group": sender_group,
            "receiver": self.receiver.to_dict(),
            "timestamp": self.timestamp,
        }


def from_dict(data: Dict[str, str or float or Dict[str, str or int]])\
        -> Message:
    """
    Generates a Message object from a dictionary

    :param data: The dictionary to turn into a Message
    :return: The generates Message object
    """

    if data["sender_group"] is None:
        sender_group = None
    else:
        sender_group = contact_from_dict(data["sender_group"])
    return Message(
        data["message_title"],
        data["message_body"],
        contact_from_dict(data["receiver"]),
        contact_from_dict(data["sender"]),
        sender_group,
        data["timestamp"]
    )



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import re
import sys
from kudubot.users.Contact import Contact
from kudubot.entities.Message import Message
from yowsup.layers.interface import YowInterfaceLayer
from yowsup.layers.interface import ProtocolEntityCallback
from yowsup.layers.protocol_media.mediauploader import MediaUploader
from yowsup.layers.protocol_messages.protocolentities import \
    MessageProtocolEntity, TextMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities import \
    RequestUploadIqProtocolEntity, ResultRequestUploadIqProtocolEntity, \
    VideoDownloadableMediaMessageProtocolEntity, \
    AudioDownloadableMediaMessageProtocolEntity, \
    ImageDownloadableMediaMessageProtocolEntity
from yowsup.layers.protocol_media.protocolentities.\
    builder_message_media_downloadable import DownloadableMediaMessageBuilder


class EchoLayer(YowInterfaceLayer):
    """
    The Yowsup Echo Layer used to communicate with the Whatsapp Servers
    """

    # connection: kudubot_whatsapp.WhatsappConnection.WhatsappConnection
    def __init__(self, connection):
        """
        Custom constructor used to pass the WhatsappConnection
        class to the layer

        :param connection: The Whatsapp connection object that calls this layer
        """
        # Required local variables
        self.disconnect_action_prompt = 0
        self.accountDelWarnings = 0
        self.connected = False
        self.username = None
        self.sendReceipts = True
        self.disconnectAction = 0
        self.credentials = None
        self.jid_aliases = {}

        super().__init__()
        self.connection = connection

    def send_text_message(self, body: str, receiver: str):
        """
        Sends a text message

        :param body: The message body
        :param receiver: The recipient address
        :return: None
        """
        self.toLower(TextMessageProtocolEntity(body, to=receiver))

    def send_image_message(self, image_file: str, receiver: str, caption: str):
        """
        Sends an image file

        :param image_file: The image file to send
        :param receiver: The recipient of the file
        :param caption: A caption to be displayed with the image
        :return: None
        """
        self.send_media(image_file, receiver, caption,
                        RequestUploadIqProtocolEntity.MEDIA_TYPE_IMAGE)

    def send_audio_message(self, audio_file: str, receiver: str, caption: str):
        """
        Sends an audio file

        :param audio_file: The audio file to send
        :param receiver: The recipient of the file
        :param caption: A caption to be displayed with the audio file
        :return: None
        """
        self.send_media(audio_file, receiver, caption,
                        RequestUploadIqProtocolEntity.MEDIA_TYPE_AUDIO)

    def send_video_message(self, video_file: str, receiver: str, caption: str):
        """
        Sends a video file

        :param video_file: The video file to send
        :param receiver: The recipient of the file
        :param caption: A caption to be displayed with the video file
        :return: None
        """
        self.send_media(video_file, receiver, caption,
                        RequestUploadIqProtocolEntity.MEDIA_TYPE_VIDEO)

    @ProtocolEntityCallback("receipt")
    def on_receipt(self, entity: MessageProtocolEntity):
        """
        Method run whenever a whatsapp receipt is issued
        This ensures that a message is marked as 'read'

        :param entity: The receipt entity
        :return: None
        """
        self.toLower(entity.ack())

    def send_receipt(self, message_protocol_entity: MessageProtocolEntity):
        """
        Sends the whatsapp servers that the message was received

        :param message_protocol_entity: the message protocol entity
                                        that was received
        :return: None
        """
        self.toLower(message_protocol_entity.ack())
        self.toLower(message_protocol_entity.ack(True))

    @ProtocolEntityCallback("message")
    def on_message(self, message_protocol_entity: TextMessageProtocolEntity):
        """
        Method run when a message is received
        :param message_protocol_entity: the message received
        :return: void
        """
        # Send receipt
        self.toLower(message_protocol_entity.ack())
        self.toLower(message_protocol_entity.ack(True))

        # Check for the message type
        if message_protocol_entity.getType() == "text":

            body = message_protocol_entity.getBody()
            timestamp = float(message_protocol_entity.getTimestamp())

            if re.search(
                    r"[0-9]+-[0-9]+",
                    message_protocol_entity.getFrom(False)
            ):  # If group
                sender = Contact(-1, message_protocol_entity.getNotify(),
                                 message_protocol_entity.getParticipant(True))
                group = Contact(-1, message_protocol_entity.getNotify(),
                                message_protocol_entity.getFrom(True))
            else:
                sender = Contact(-1, message_protocol_entity.getNotify(),
                                 message_protocol_entity.getFrom(True))
                group = None

            message = Message("", body, self.connection.user_contact,
                              sender, group, timestamp)
            self.connection.apply_services(message, True)

    @staticmethod
    def normalize_jid(number: str) -> str:
        """
        Normalizes a jid

        :param number: the number to be receive a normalized jid
        :return: the normalized jid
        """
        if '@' in number:
            return number
        elif "-" in number:
            return number + "@g.us"

        return number + "@s.whatsapp.net"

    def alias_to_jid(self, c_alias: str) -> str:
        """
        Turns an alias into a jid

        :param c_alias: the current alias
        :return: the jid
        """
        for alias, a_jid in self.jid_aliases.items():
            if c_alias.lower() == alias.lower():
                return EchoLayer.normalize_jid(a_jid)

        return EchoLayer.normalize_jid(c_alias)

    def send_media(self, path: str, number: str, caption: str,
                   media_type: RequestUploadIqProtocolEntity.TYPES_MEDIA):
        """
        Sends a media file

        :param path: the path to the media file
        :param number: the receiver of the media file
        :param caption: the caption to be shown
        :param media_type: The type of media to send
        :return: None
        """
        jid = self.alias_to_jid(number)
        entity = RequestUploadIqProtocolEntity(media_type, filePath=path)

        def success_fn(success_entity: TextMessageProtocolEntity,
                       original_entity: MessageProtocolEntity):
            """
            Function called on successful media upload

            :param success_entity: the success entity
            :param original_entity: the original entity
            :return: None
            """
            # noinspection PyTypeChecker
            self.on_request_upload_result(jid, media_type, path,
                                          success_entity, original_entity,
                                          caption)

        def error_fn(error_entity: MessageProtocolEntity,
                     original_entity: RequestUploadIqProtocolEntity):
            """
            Function called on failed media upload

            :param error_entity: the error entity
            :param original_entity: the original entity
            :return: None
            """
            EchoLayer.on_request_upload_error(
                jid, path, error_entity, original_entity
            )

        self._sendIq(entity, success_fn, error_fn)

    def do_send_media(self,
                      media_type: RequestUploadIqProtocolEntity.TYPES_MEDIA,
                      file_path: str, url: str, to: str, ip: str = None,
                      caption: str = None):
        """
        Sends a media file

        :param media_type: The type of media to send
        :param file_path: the path to the file
        :param url: the whatsapp upload url
        :param to: the receiver
        :param ip: the ip of the receiver
        :param caption: the caption to be displayed together
                        with the media file
        :return: None
        """
        entity = None
        if media_type == RequestUploadIqProtocolEntity.MEDIA_TYPE_IMAGE:
            entity = ImageDownloadableMediaMessageProtocolEntity.fromFilePath(
                file_path, url, ip, to, caption=caption)
        elif media_type == RequestUploadIqProtocolEntity.MEDIA_TYPE_AUDIO:
            entity = self.create_audio_media(file_path, url, to, ip)
        elif media_type == RequestUploadIqProtocolEntity.MEDIA_TYPE_VIDEO:
            entity = VideoDownloadableMediaMessageProtocolEntity.fromFilePath(
                file_path, url, ip, to, caption=caption)
        self.toLower(entity)

    @staticmethod
    def create_audio_media(file_path: str, url: str, to: str, ip: str = None) \
            -> AudioDownloadableMediaMessageProtocolEntity:
        """
        Creates an Audio Message Entity as a workaround for the missing
        fromFilePath method in
        AudioDownloadableMediaMessageProtocolEntity's parent class

        :param file_path: the path to the file
        :param url: the whatsapp upload url
        :param to: the receiver
        :param ip: the ip of the receiver
        :return: None
        :return: the generatd audio entity
        """
        builder = DownloadableMediaMessageBuilder(
            AudioDownloadableMediaMessageProtocolEntity, to, file_path)
        builder.set("url", url)
        builder.set("ip", ip)
        return AudioDownloadableMediaMessageProtocolEntity.fromBuilder(builder)

    def on_request_upload_result(self, jid: str,
                                 media_type:
                                     RequestUploadIqProtocolEntity.TYPES_MEDIA,
                                 file_path: str,
                                 result_request_upload_iq_protocol_entity:
                                     ResultRequestUploadIqProtocolEntity,
                                 request_upload_iq_protocol_entity:
                                     RequestUploadIqProtocolEntity,
                                 caption: str = None):
        """
        Method run when a media upload result is positive

        :param jid: the jid to receive the media
        :param media_type: The media type to send
        :param file_path: the path to the media file
        :param result_request_upload_iq_protocol_entity: the result entity
        :param request_upload_iq_protocol_entity: the request entity
        :param caption: the media caption, if applicable
        :return: None
        """
        str(request_upload_iq_protocol_entity)
        if result_request_upload_iq_protocol_entity.isDuplicate():
            self.do_send_media(
                media_type, file_path,
                result_request_upload_iq_protocol_entity.getUrl(),
                jid, result_request_upload_iq_protocol_entity.getIp(), caption)
        else:

            def success_fn(inner_file_path: str, inner_jid: str, url: str):
                """
                Function called when upload was successful

                :param inner_file_path: path to the media file
                :param inner_jid: the receiver's jid
                :param url: the whatsapp media url
                :return: None
                """
                self.do_send_media(
                    media_type, inner_file_path, url, inner_jid,
                    result_request_upload_iq_protocol_entity.getIp(), caption)

            media_uploader = MediaUploader(
                jid, self.getOwnJid(), file_path,
                result_request_upload_iq_protocol_entity.getUrl(),
                result_request_upload_iq_protocol_entity.getResumeOffset(),
                success_fn,
                EchoLayer.on_upload_error, EchoLayer.on_upload_progress
                # async=False  Reserved keyword in python3.7+
            )
            media_uploader.start()

    @staticmethod
    def on_request_upload_error(jid: str, path: str,
                                error_request_upload_iq_protocol_entity:
                                    MessageProtocolEntity,
                                request_upload_iq_protocol_entity:
                                    RequestUploadIqProtocolEntity):
        """
        Method run when a media upload result is negative

        :param jid: the jid to receive the media
        :param path: the file path to the media
        :param error_request_upload_iq_protocol_entity: the error result entity
        :param request_upload_iq_protocol_entity: the request entity
        :return: None
        """
        if error_request_upload_iq_protocol_entity \
                and request_upload_iq_protocol_entity:
            str(jid + path)

    @staticmethod
    def on_upload_error(file_path: str, jid: str, url: str):
        """
        Method run when an upload error occurs

        :param file_path: the file path of the file to upload
        :param jid: the jid of the receiver of the file
        :param url: the upload url
        :return: None
        """
        pass

    @staticmethod
    def on_upload_progress(file_path: str, jid: str, url: str,
                           progress: float):
        """
        Method that keeps track of the upload process

        :param file_path: the file path of the media file
        :param jid: the jid of the receiver
        :param url: the whatsapp upload url
        :param progress: the current progress
        :return:None
        """
        if url:
            sys.stdout.write("%s => %s, %d%% \r" % (
                os.path.basename(file_path), jid, progress))
            sys.stdout.flush()



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import unittest
from kudubot.services.BaseService import BaseService
from kudubot.tests.helpers.DummyService import DummyService


class UnitTests(unittest.TestCase):
    """
    Tests the Service class
    """

    def setUp(self):
        """
        :return: None
        """
        pass

    def tearDown(self):
        """
        :return: None
        """
        pass

    def test_abstract_methods(self):
        """
        Tests if the methods of the Service class are abstract
        :return: None
        """
        dummy = DummyService([])

        for method in [(BaseService.handle_message, 1),
                       (BaseService.is_applicable_to, 1),
                       (BaseService.define_identifier, -1)]:
            try:

                if method[1] == -1:
                    method[0]()
                elif method[1] == 1:
                    method[0](dummy, dummy)
                self.fail()
            except NotImplementedError:
                pass



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import configparser
from typing import Dict, List
from yowsup.layers import YowLayerEvent
from yowsup.stacks import YowStackBuilder
from kudubot.users.Contact import Contact
from kudubot.entities.Message import Message
from yowsup.layers.network import YowNetworkLayer
from kudubot.connections.Connection import Connection
from kudubot.exceptions import InvalidConfigException
from kudubot.connections.whatsapp.EchoLayer import EchoLayer
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


class WhatsappConnection(Connection):
    """
    Class that implements a kudubot connection for
    the Whatsapp Messaging Service
    """

    def __init__(self, services: List[type],
                 config_handler: GlobalConfigHandler):
        """
        Extends the default Connection constructor to create a yowsup stack

        :param services: The services to start
        :param config_handler: The GlobalConfigHandler that determines the
                               location of the configuration files
        """
        super().__init__(services, config_handler)

        stack_builder = YowStackBuilder()
        self.yowsup = EchoLayer(self)
        self.stack = \
            stack_builder.pushDefaultLayers(True).push(self.yowsup).build()
        self.stack.setCredentials((self.config["number"], self.config["pass"]))

    @staticmethod
    def define_identifier() -> str:
        """
        Defines a unique identifier for the connection

        :return: 'whatsapp'
        """
        return "whatsapp"

    def define_user_contact(self) -> Contact:
        """
        :return: The Whatsapp connection's contact information
        """
        return Contact(-1, "Kudubot", self.config["number"])

    def generate_configuration(self):
        """
        Generates a new Configuration file for the Whatsapp Connection

        :return: None
        """
        with open(self.config_file_location, 'w') as config:
            config.write(
                "[credentials]\nnumber=YourNumberHere\npass=YourPassHere\n")
            self.logger.info(
                "Wrote new Configuration file at " + self.config_file_location)

    def load_config(self) -> Dict[str, str]:
        """
        Loads the configuration for the Whatsapp Connection
        from the config file

        :return: The parsed configuration,
                 consisting of a dictionary with a number and pass key
        """
        try:

            self.logger.info("Parsing config at " + self.config_file_location)

            config = configparser.ConfigParser()
            config.read(self.config_file_location)
            parsed_config = dict(config.items("credentials"))

            if parsed_config["number"] == "":
                self.logger.warning(
                    "Config Parsing Failed. No number supplied.")
                raise InvalidConfigException(
                    "Invalid Whatsapp Configuration File - "
                    "Missing number detected")
            elif parsed_config["pass"] == "":
                self.logger.warning(
                    "Config Parsing Failed. No password supplied.")
                raise InvalidConfigException(
                    "Invalid Whatsapp Configuration File - "
                    "Missing password detected")

            self.logger.info("Config successfully loaded")

            return parsed_config

        except (configparser.NoSectionError,
                configparser.MissingSectionHeaderError):

            self.logger.warning("Config Parsing Failed. "
                                "No credentials section in config file.")
            raise InvalidConfigException("Invalid Whatsapp Configuration File "
                                         "- No credentials section")
        except KeyError as e:
            self.logger.warning(
                "Config Parsing Failed. No attribute " + str(e) + " found.")
            raise InvalidConfigException(
                "Invalid Whatsapp Configuration File - No '" + str(e) +
                "' attribute found")

    def send_message(self, message: Message):
        """
        Sends a Text message using the Whatsapp Connection

        :param message: The message to send
        :return: None
        """
        self.yowsup.send_text_message(
            message.message_body, message.receiver.address)

    def send_audio_message(self, receiver: Contact, audio_file: str,
                           caption: str = ""):
        """
        Sends an audio message using the Whatsapp Connection

        :param receiver: The receiver of the audio message
        :param audio_file: The audio file to send
        :param caption: An optional caption for the audio message
        :return: None
        """
        self.yowsup.send_audio_message(audio_file, receiver.address, caption)

    def send_video_message(self, receiver: Contact, video_file: str,
                           caption: str = ""):
        """
        Sends a video message using the Whatsapp Connection

        :param receiver: The receiver of the video message
        :param video_file: The video file to send
        :param caption: An optional caption for the video message
        :return: None
        """
        self.yowsup.send_video_message(video_file, receiver.address, caption)

    def send_image_message(self, receiver: Contact, image_file: str,
                           caption: str = ""):
        """
        Sends an image message using the Whatsapp Connection

        :param receiver: The receiver of the image message
        :param image_file: The image file to send
        :param caption: An optional caption for the image message
        :return: None
        """
        self.yowsup.send_image_message(image_file, receiver.address, caption)

    def listen(self):
        """
        Starts the Yowsup listener

        :return: None
        """
        self.stack.broadcastEvent(
            YowLayerEvent(YowNetworkLayer.EVENT_STATE_CONNECT)
        )
        self.stack.loop()



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from kudubot.entities.Message import Message
from kudubot.services.BaseService import BaseService


class EchoService(BaseService):
    """
    Service that always replies to a user with the same message body that was
    received
    """

    def is_applicable_to(self, message: Message) -> bool:
        """
        This service is always applicable
        :param message: The message to check for applicability for
        :return: True
        """
        return True

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the identifier of the service
        :return: "echo"
        """
        return "echo"

    def handle_message(self, message: Message):
        """
        Replies with the original message text
        :param message: The original message
        :return: None
        """
        self.reply("Echo", message.message_body, message)



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import unittest
from kudubot.users.Contact import Contact


class UnitTests(unittest.TestCase):
    """
    Tests the Contact class
    """

    def setUp(self):
        """
        :return: None
        """
        pass

    def tearDown(self):
        """
        :return: None
        """
        pass

    def test_initialization(self):
        """
        Tests if the Contact class is initialized correctly

        :return: None
        """
        contact = Contact(1, "A", "B")
        self.assertEqual(contact.database_id, 1)
        self.assertEqual(contact.display_name, "A")
        self.assertEqual(contact.address, "B")



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import unittest

from kudubot.entities.Message import Message
from kudubot.users.Contact import Contact


class UnitTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_creating_message(self):

        message = Message("title", "body", Contact(-1, "A", "Add"),
                          Contact(-2, "B", "Badd"), Contact(-3, "C", "Cadd"),
                          timestamp=1.0)

        self.assertEqual(message.message_title, "title")
        self.assertEqual(message.message_body, "body")
        self.assertEqual(message.receiver.display_name, "A")
        self.assertEqual(message.sender.display_name, "B")
        self.assertEqual(message.sender_group.display_name, "C")
        self.assertEqual(message.timestamp, 1.0)



"""LICENSE
Copyright 2015 Hermann Krumrey <hermann@krumreyh.com>

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import List

from kudubot.entities.Message import Message
from kudubot.services.BaseService import BaseService


class DummyService(BaseService):
    """
    A class that implements a Service for use in unit tests
    """

    @staticmethod
    def define_requirements() -> List[str]:
        return []

    @staticmethod
    def define_identifier() -> str:
        return "dummyservice"

    def handle_message(self, message: Message):
        pass

    def is_applicable_to(self, message: Message) -> bool:
        pass


class DummyServiceWithValidDependency(DummyService):
    """
    A Service that has itself (or DummyService) as its only dependency
    """

    @staticmethod
    def define_requirements() -> List[str]:
        return ["dummyservice"]


class DummyServiceWithInvalidDependency(DummyService):
    """
    A service that has an invalid dependency
    """

    @staticmethod
    def define_requirements() -> List[str]:
        return ["otherservice"]
